# Simple input/output using MessagePack and LZ4 for Python
# Copyright 2019 Hugo Hromic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for data input/output."""

from contextlib import contextmanager
from .compression import decompressor, compressor
from .packing import unpacker, packer

@contextmanager
def data_unpacker(path, compression=True, **kwargs):
    """Create a data unpacker (MessagePack) context manager with
       optional compression (LZ4) support."""
    with decompressor(path) if compression else open(path, "rb") as reader:
        yield unpacker(reader, **kwargs)

@contextmanager
def data_pack(path, compression=True, level=None, append=False, **kwargs):
    """Create a data pack (MessagePack) context manager with
       optional compression (LZ4) support and file appending."""
    if compression:
        context = compressor(path, level=level, append=append)
    else:
        context = open(path, "ab" if append else "wb")
    with context as writer:
        pkr = packer(**kwargs)
        def _pack(obj):
            writer.write(pkr.pack(obj))
        yield _pack

def load_obj(path, **kwargs):
    """Load an object from disk."""
    with data_unpacker(path, **kwargs) as _unpacker:
        return next(_unpacker)

def load_iter(path, **kwargs):
    """Iterate objects from disk (MessagePack-LZ4 format)."""
    with data_unpacker(path, **kwargs) as _unpacker:
        for obj in _unpacker:
            yield obj

def save_obj(obj, path, **kwargs):
    """Save an object to disk (MessagePack-LZ4 format)."""
    with data_pack(path, **kwargs) as _pack:
        _pack(obj)

def save_iter(iterable, path, **kwargs):
    """Save an iterable to disk (MessagePack-LZ4 format)."""
    with data_pack(path, **kwargs) as _pack:
        for element in iterable:
            _pack(element)



# Simple input/output using MessagePack and LZ4 for Python
# Copyright 2019 Hugo Hromic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Encoders/decoders for Python types to be used with MessagePack (https://msgpack.org/)."""

from datetime import datetime

def encode_set(obj, typecode, packb, ext_type):
    """Encode a Python 'set' object into a MessagePack ExtType."""
    return ext_type(typecode, packb(tuple(obj)))

def decode_set(data, unpackb):
    """Decode MessagePack data into a Python 'set' object."""
    return set(unpackb(data, use_list=False))

def encode_datetime(obj, typecode, packb, ext_type):
    """Encode a Python 'datetime' object into a MessagePack ExtType."""
    delta = obj - datetime(1970, 1, 1, tzinfo=obj.tzinfo)
    data = delta.seconds + delta.days * 24 * 3600, delta.microseconds
    return ext_type(typecode, packb(data))

def decode_datetime(data, unpackb):
    """Decode MessagePack data into a Python 'datetime' object."""
    seconds, microseconds = unpackb(data)
    return datetime.utcfromtimestamp(seconds).replace(microsecond=microseconds)



# Simple input/output using MessagePack and LZ4 for Python
# Copyright 2019 Hugo Hromic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Command-line tools."""

import sys
from argparse import ArgumentParser
from pprint import PrettyPrinter
from .inpout import load_iter

def pprint():
    """Pretty-print files in MsgPack format with optional compression."""
    parser = ArgumentParser(description=pprint.__doc__)
    parser.add_argument("filenames", metavar="FILENAME", nargs="+",
                        help="an input filename to pretty-print")
    parser.add_argument("-n", metavar="NUMBER", type=int,
                        help="only process the first NUMBER objects"
                             " from each input file")
    args = parser.parse_args()

    # read and pretty-print the given input files
    pprinter = PrettyPrinter()
    for fname in args.filenames:
        sys.stderr.write("Processing file '%s' ...\n" % fname)
        try:
            compression = fname.lower().endswith(".lz4")
            for cnt, obj in enumerate(load_iter(fname, compression=compression, use_list=False)):
                if args.n is None or cnt < args.n:
                    pprinter.pprint(obj)
        except Exception as excp:  # pylint: disable=broad-except
            sys.stderr.write("%s\n" % excp)



# Simple input/output using MessagePack and LZ4 for Python
# Copyright 2019 Hugo Hromic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Simple input/output using MessagePack and LZ4 package."""

from datetime import datetime
from functools import partial
from . import compression
from . import packing
from . import ext_types
from .inpout import (
    data_unpacker,
    data_pack,
    load_obj,
    load_iter,
    save_obj,
    save_iter,
)

# register included extended type for Python 'set' objects
packing.register_ext_type_encoder(set, partial(ext_types.encode_set, typecode=127))
packing.register_ext_type_decoder(127, ext_types.decode_set)

# register included extended type for Python 'datetime' objects
packing.register_ext_type_encoder(datetime, partial(ext_types.encode_datetime, typecode=126))
packing.register_ext_type_decoder(126, ext_types.decode_datetime)



# Simple input/output using MessagePack and LZ4 for Python
# Copyright 2019 Hugo Hromic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Functions for data packing and unpacking using MessagePack (https://msgpack.org/)."""

import msgpack

# global extended type encoders mapping (class -> encoder)
_EXT_TYPE_ENCODERS = {}

# global extended type decoders mapping (typecode -> decoder)
_EXT_TYPE_DECODERS = {}

def _encode_ext_type(obj):
    obj_type = type(obj)
    if obj_type in _EXT_TYPE_ENCODERS:
        return _EXT_TYPE_ENCODERS[obj_type](obj, packb=packb, ext_type=msgpack.ExtType)
    raise TypeError("unknown extended type encoder for %r" % obj_type)

def _decode_ext_type(typecode, data):
    if typecode in _EXT_TYPE_DECODERS:
        return _EXT_TYPE_DECODERS[typecode](data, unpackb=unpackb)
    raise TypeError("unknown extended type decoder for typecode %d" % typecode)

def register_ext_type_encoder(python_type, encoder):
    """Register a MessagePack ExtType encoder for a Python type."""
    _EXT_TYPE_ENCODERS[python_type] = encoder

def register_ext_type_decoder(typecode, decoder):
    """Register a Python type decoder for a MessagePack ExtType typecode."""
    _EXT_TYPE_DECODERS[typecode] = decoder

def pack(obj, stream, **kwargs):
    """Pack object using MessagePack (with extended types support)
       and write packed bytes to a stream."""
    msgpack.pack(obj, stream, default=_encode_ext_type, use_bin_type=True, **kwargs)

def packb(obj, **kwargs):
    """Pack object using MessagePack (with extended types support)
       and return packed bytes."""
    return msgpack.packb(obj, default=_encode_ext_type, use_bin_type=True, **kwargs)

def packer(*args, **kwargs):
    """Return a MessagePack (with extended types support) Packer object."""
    return msgpack.Packer(*args, default=_encode_ext_type, use_bin_type=True, **kwargs)

def unpack(stream, **kwargs):
    """Unpack a stream of packed bytes using MessagePack (with extended types support)
       and return unpacked object."""
    return msgpack.unpack(stream, ext_hook=_decode_ext_type, raw=False, **kwargs)

def unpackb(packed, **kwargs):
    """Unpack packed bytes using MessagePack (with extended types support)
       and return unpacked object."""
    return msgpack.unpackb(packed, ext_hook=_decode_ext_type, raw=False, **kwargs)

def unpacker(*args, **kwargs):
    """Return a MessagePack (with extended types support) Unpacker object."""
    return msgpack.Unpacker(*args, ext_hook=_decode_ext_type, raw=False, **kwargs)



"""Main setup script."""

from setuptools import setup, find_packages

NAME = "inpout"
VERSION = "1.0.8"
DESCRIPTION = "Simple input/output using MessagePack and LZ4 for Python"
AUTHOR = "Hugo Hromic"
AUTHOR_EMAIL = "hhromic@gmail.com"
URL = "https://github.com/hhromic/python-inpout"
DOWNLOAD_URL = URL + "/tarball/" + VERSION
REQUIRES = [
    "msgpack",
    "py-lz4framed",
]
CLASSIFIERS = [
    "Topic :: System :: Archiving",
    "Intended Audience :: Developers",
]
KEYWORDS = [
    "messagepack",
    "lz4",
    "input",
    "output",
    "library",
]
ENTRY_POINTS = {
    "console_scripts": [
        "inpout-pprint = inpout.cli:pprint",
    ],
}
LICENSE = "Apache-2.0"

def _read_file(filename):
    with open(filename) as reader:
        return reader.read()

setup(
    name=NAME, version=VERSION, description=DESCRIPTION,
    author=AUTHOR, author_email=AUTHOR_EMAIL,
    maintainer=AUTHOR, maintainer_email=AUTHOR_EMAIL,
    url=URL, download_url=DOWNLOAD_URL,
    classifiers=CLASSIFIERS,
    install_requires=REQUIRES,
    provides=[NAME],
    entry_points=ENTRY_POINTS,
    keywords=KEYWORDS,
    license=LICENSE,
    platforms=["all"],
    long_description=_read_file("README.md"),
    long_description_content_type="text/markdown",
    packages=find_packages(),
)



# Simple input/output using MessagePack and LZ4 for Python
# Copyright 2019 Hugo Hromic
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Classes and functions for data compression/decompression using LZ4 (http://www.lz4.org)."""

from contextlib import contextmanager
from lz4framed import Compressor, Decompressor
from lz4framed import Lz4FramedError, Lz4FramedNoDataError
from lz4framed import LZ4F_COMPRESSION_MAX

class _WritableCompressor(Compressor):
    def write(self, data): # pylint: disable=missing-docstring
        return self.update(data)

class _ReadableDecompressor(Decompressor):
    def readall(self):  # pylint: disable=missing-docstring
        data_buffer = bytearray()
        for chunk in self:
            data_buffer += chunk
        return bytes(data_buffer)

    def read(self, n=-1):  # pylint: disable=missing-docstring, invalid-name, unused-argument
        if n < 0:
            return self.readall()
        if n == 0:
            return bytes()
        iterator = iter(self)
        data_buffer = bytearray(n)
        bytes_read = 0
        while bytes_read < n:
            try:
                chunk = next(iterator)
                data_buffer[bytes_read:len(chunk)] = chunk
                bytes_read += len(chunk)
            except (Lz4FramedError, Lz4FramedNoDataError, StopIteration):
                break
        return bytes(data_buffer[0:bytes_read])

@contextmanager
def compressor(path, level=None, append=False):
    """Create a data compressing context manager for file writing.
       When 'level' is None, a default LZ4F_COMPRESSION_MAX level is used.
       When 'append' is True, the open mode is set for appending data."""
    level = LZ4F_COMPRESSION_MAX if level is None else level
    with open(path, "ab" if append else "wb") as writer:
        with _WritableCompressor(writer, level=level) as context:
            yield context

@contextmanager
def decompressor(path):
    """Create a data decompressing context manager for file reading."""
    with open(path, "rb") as reader:
        yield _ReadableDecompressor(reader)
